/* Connect.cpp - MyopicTopics.com Addin - part 4
 * Written By: Jason Weiler
 * This code is in the public domain
 */

// Connect.cpp : Implementation of CConnect
#include "stdafx.h"
#include "AddIn.h"
#include "Connect.h"

#include "DebuggerData.h"

extern CAddInModule _AtlModule;
DebuggerData CConnect::s_DebuggerData;

// Evaluate once per sesssion
CComBSTR const bsSizeofPointer(L"sizeof(void*),d");
CComBSTR const bsSizeofInteger(L"sizeof(int),d");
CComBSTR const bsOffsetToStringIndex(L"(int)&(((MyType*)0)->StringIndex),d");

// Evaluate once per break
CComBSTR const bsGlobalStringTableData(L"(int)GlobalStringTable,d");
CComBSTR const bsGlobalStringTableMax(L"(int)GlobalStringTableMax,d");


// CConnect
STDMETHODIMP CConnect::OnConnection(IDispatch *pApplication, ext_ConnectMode ConnectMode, IDispatch *pAddInInst, SAFEARRAY ** /*custom*/ )
{
	// unused variables
	(void)ConnectMode;

	HRESULT hr = S_OK;
	pApplication->QueryInterface(__uuidof(DTE2), (LPVOID*)&m_pDTE);
	pAddInInst->QueryInterface(__uuidof(AddIn), (LPVOID*)&m_pAddInInstance);

	if ( !m_pDTE || !m_pAddInInstance)
	{
		// This should never happen, but we don't want this add-in running if it does.
		return E_FAIL;
	}

//*** Hooking debugger events ***/
	// Register the OnEnterBreakMode handler
	if ( !m_pDebuggerEvents )
	{
		CComPtr<EnvDTE::Events> spEvents;
		m_pDTE->get_Events(&spEvents);
		if (spEvents->get_DebuggerEvents(&m_pDebuggerEvents) == S_OK)
		{
			// Attach to the debugger events
			hr = ConnectDebuggerEvents::DispEventAdvise(m_pDebuggerEvents);
		}
	}
//*** Hooking debugger events ***/

	return hr;
}

STDMETHODIMP CConnect::OnDisconnection(ext_DisconnectMode /*RemoveMode*/, SAFEARRAY ** /*custom*/ )
{
//*** Hooking debugger events ***/
	// Register the OnEnterBreakMode handler
	if (m_pDebuggerEvents)
	{
		// Let go of the debugger events
		ConnectDebuggerEvents::DispEventUnadvise(m_pDebuggerEvents);
	}
//*** Hooking debugger events ***/

	// Explicitly set these to NULL because our object may not be completely going away.
	// These are CComPtr template types, so setting them to NULL calls Release automatically.
	m_pDebuggerEvents = NULL;
	m_pDTE = NULL;
	m_pAddInInstance = NULL;

	return S_OK;
}

STDMETHODIMP CConnect::OnAddInsUpdate (SAFEARRAY ** /*custom*/ )
{
	return S_OK;
}

STDMETHODIMP CConnect::OnStartupComplete (SAFEARRAY ** /*custom*/ )
{
	return S_OK;
}

STDMETHODIMP CConnect::OnBeginShutdown (SAFEARRAY ** /*custom*/ )
{
	return S_OK;
}

//*** Hooking debugger events ***/
STDMETHODIMP CConnect::OnEnterBreakMode(EnvDTE::dbgEventReason Reason, EnvDTE::dbgExecutionAction* ExecutionAction)
{
	// unused variables
	(void)ExecutionAction;
	(void)Reason;

	if (!m_pDTE)
	{
		// this shouldn't happen
		return S_OK;
	}

	HRESULT hr = S_OK;


	long lPID = -1;

	CComPtr<EnvDTE::Debugger> pDebugger = NULL;
	hr = m_pDTE->get_Debugger(&pDebugger);
	if (hr != S_OK)
	{
		// no debugger? This shouldn't happen.
		return S_OK;
	}

	// assume we're going to query everything
	bool bUpdatePerSessionData = true;

	if (GetCurrentPID(pDebugger, &lPID))
	{
		if (s_DebuggerData.lCurrentPID != -1
			&& s_DebuggerData.lCurrentPID == lPID)
		{
			// We're in the same session as last time, so skip the per-session queries
			bUpdatePerSessionData = false;
		}
		else
		{
			// This looks like a new session, so clear out the cached data.
			s_DebuggerData.ClearData();
		}
	}

	// Get a copy of the static debugger data
	DebuggerData tmpDebuggerData = s_DebuggerData;

	CComBSTR bsValue = NULL;
	bool bAllQueriesSuccessful = true;
	
	if (bUpdatePerSessionData)
	{
		tmpDebuggerData.lCurrentPID = lPID;

		// How large is a pointer on this target platform?
		if ( QueryExpression(pDebugger, bsSizeofPointer, &bsValue) && bsValue.Length() )
		{
			tmpDebuggerData.SizeOfPointer = wcstol(bsValue, NULL, 10);
		}
		else
		{
			bAllQueriesSuccessful = false;
		}
		bsValue.Empty();

		// How large is a pointer on this target platform?
		if ( QueryExpression(pDebugger, bsSizeofInteger, &bsValue) && bsValue.Length() )
		{
			tmpDebuggerData.SizeOfInteger = wcstol(bsValue, NULL, 10);
		}
		else
		{
			bAllQueriesSuccessful = false;
		}
		bsValue.Empty();


		// What is the offset to the StringIndex member?
		if ( QueryExpression(pDebugger, bsOffsetToStringIndex, &bsValue) && bsValue.Length() )
		{
			tmpDebuggerData.MyTypeOffsetToStringIndex = wcstol(bsValue, NULL, 10);
		}
		else
		{
			bAllQueriesSuccessful = false;
		}
		bsValue.Empty();
	}

	// What is the current value of the GlobalStringTable pointer?
	if ( QueryExpression(pDebugger, bsGlobalStringTableData, &bsValue) && bsValue.Length() )
	{
		// We should know the size of a pointer by now, so we need to make
		//   sure we treat this data appropriately
		if (tmpDebuggerData.SizeOfPointer == 8)
		{
			tmpDebuggerData.GlobalStringTableData = _wcstoui64(bsValue, NULL, 10);
		}
		else if (tmpDebuggerData.SizeOfPointer == 4)
		{
			tmpDebuggerData.GlobalStringTableData = ((unsigned __int64)wcstoul(bsValue, NULL, 10)) & 0x00000000FFFFFFFF;
		}
		bsValue.Empty();

	}
	else
	{
		bAllQueriesSuccessful = false;
	}

	// What is the current max-size of the GlobalStringTable pointer?
	if ( QueryExpression(pDebugger, bsGlobalStringTableMax, &bsValue) && bsValue.Length() )
	{
		// We should know the size of a pointer by now, so we need to make
		//   sure we treat this data appropriately
		if (tmpDebuggerData.SizeOfPointer == 8)
		{
			tmpDebuggerData.GlobalStringTableMax = _wcstoui64(bsValue, NULL, 10);
		}
		else if (tmpDebuggerData.SizeOfPointer == 4)
		{
			tmpDebuggerData.GlobalStringTableMax = ((unsigned __int64)wcstoul(bsValue, NULL, 10)) & 0x00000000FFFFFFFF;
		}
		bsValue.Empty();

	}
	else
	{
		bAllQueriesSuccessful = false;
	}


	// If and ONLY if all our queries worked do we want to write this data back to the cache.
	//   We don't want to get into the situation where only some data is valid.
	if (bAllQueriesSuccessful)
	{
		s_DebuggerData = tmpDebuggerData;
	}

	return S_OK;
}
//*** Hooking debugger events ***/

bool CConnect::QueryExpression(EnvDTE::Debugger * pDebugger
								, CComBSTR const & bsExpression
								, BSTR * bsOutValue)
{
	if (NULL == pDebugger)
		return false;

	bool bRetval = true;

	CComPtr<EnvDTE::Expression> pExpr = NULL;
	HRESULT hr = pDebugger->GetExpression(bsExpression, VARIANT_FALSE, -1, &pExpr);
	if (hr == S_OK)
	{

		VARIANT_BOOL bIsValid=true;
		hr = pExpr->get_IsValidValue(&bIsValid);
		if (hr == S_OK || !bIsValid)
		{
			pExpr->get_Value(bsOutValue);

			if (!bsOutValue)
			{
				bRetval = false;
			}
		}
		else
		{
			bRetval = false;
		}
	}

	return bRetval;
}

bool CConnect::GetCurrentPID(EnvDTE::Debugger * pDebugger, long * lOutPID)
{
	CComPtr<EnvDTE::Process> pProcess = NULL;
	HRESULT hr = pDebugger->get_CurrentProcess(&pProcess);
	if (hr == S_OK)
	{
		hr = pProcess->get_ProcessID(lOutPID);
		if (hr == S_OK)
		{
			return true;
		}
	}

	*lOutPID = -1;
	return false;
}

DebuggerData const & GetStaticDebuggerData()
{
	return CConnect::s_DebuggerData;
}
